/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package factories;

import Facemorph.MultiscaleWarp;
import MorphAnalyser.ObjectGeometry;
import MorphAnalyser.Template3D;
//import com.sun.j3d.utils.pickfast.PickTool;
import com.sun.j3d.utils.picking.*;

import java.awt.Image;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Scanner;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.imageio.ImageIO;
import javax.media.j3d.Appearance;
import javax.media.j3d.GLSLShaderProgram;
import javax.media.j3d.GeometryArray;
import javax.media.j3d.ImageComponent2D;
import javax.media.j3d.Material;
import javax.media.j3d.PolygonAttributes;
import javax.media.j3d.Shader;
import javax.media.j3d.ShaderAppearance;
import javax.media.j3d.ShaderAttribute;
import javax.media.j3d.ShaderAttributeSet;
import javax.media.j3d.ShaderAttributeValue;
import javax.media.j3d.ShaderProgram;
import javax.media.j3d.Shape3D;
import javax.media.j3d.SourceCodeShader;
import javax.media.j3d.Texture;
import javax.media.j3d.Texture2D;
import javax.media.j3d.TextureAttributes;
import javax.media.j3d.TextureUnitState;
import javax.media.j3d.TriangleArray;
import javax.swing.ImageIcon;
import javax.vecmath.Color3f;

/**
 *
 * @author bpt
 */
public class MultiTextureObjectGeometry extends ObjectGeometry {
    public boolean hasMultiText = false;
    public ArrayList<BufferedImage> textures = new ArrayList<BufferedImage>();

    public MultiTextureObjectGeometry(Template3D template, ObjectGeometry obj) {
        super(template, obj);
    }

    public MultiTextureObjectGeometry() {
        super();
    }
     /**
     * Reads a Alias wavefront .obj file or a .avrg file and the associated
     * texture map.
     *
     * @param objFile
     *            The name of the object file to load
     * @param imgFile
     *            The name of the texture map image.
     * @return true if the file was successfully read.
     */
    @Override
    public boolean read(String objFile, String imgFile) throws FileNotFoundException, IOException {
        boolean res = super.read(objFile, imgFile);
        if (!res) return res;
        String multiFileStr = objFile.substring(0, objFile.lastIndexOf("."))+".multi";
        File multiFile = new File(multiFileStr);
        textures.clear();
        if (!multiFile.exists()) {
            hasMultiText = false;
            return true;
        }
        hasMultiText = true;
        Scanner sc = new Scanner(multiFile);
        String folder = multiFile.getParent();
        while (sc.hasNextLine()) {
            String line = sc.nextLine();
            BufferedImage bimg = ImageIO.read(new File(folder + File.separator + line));
            textures.add(bimg);
        }
        return true;
    }

    @Override
    public boolean write(String objFile, String imgFile) throws IOException {
        boolean res = super.write(objFile, imgFile);
        if (!res) return res;
        if (hasMultiText) {
            String multiFileStr = objFile.substring(0, objFile.lastIndexOf("."))+".multi";
            File multiFile = new File(multiFileStr);
            PrintWriter pw = new PrintWriter(multiFile);
            for (int i=0; i<textures.size(); i++) {
                BufferedImage bimg = textures.get(i);
                String fileStr = objFile.substring(0, objFile.lastIndexOf("."))+"multi_" + i + ".jpg";
                ImageIO.write(bimg, "jpeg", new File(fileStr));
                pw.println("multi_" + i + ".jpg");
            }
        }
        return true;
    }
    /**
     * Aligns this ObjectGeometry with target using the two 3D templates to
     * provide an initial partial alignment and then using ray-tracing from
     * vertices to target triangles to improve the alignment
     *
     * @param t1
     *            This ObjectGeometry's template
     * @param t2
     *            The target ObjectGeometry's template
     * @param target
     *            The target ObjectGeometry
     */
    @Override
    public void warpICP3DtraceTriangles(Template3D t1, Template3D t2,
            ObjectGeometry target, boolean split) {
        warp3D(t1, t2);
        ArrayList[] coords2D = warpICP3DtraceTriangles(target, split, 8, t1.getMidX(), t2.getMidX(), 0);
        textures.clear();
        hasMultiText = false;
        if (target instanceof MultiTextureObjectGeometry) {
            MultiTextureObjectGeometry multiTarget = (MultiTextureObjectGeometry)target;

            if (multiTarget.hasMultiText) {
                hasMultiText = true;

                BufferedImage img = this.getImage();
                MultiscaleWarp warp = new MultiscaleWarp(img.getWidth(), img.getHeight(), target.getImage().getWidth(), target.getImage().getHeight());

                warp.interpolate(coords2D[0], coords2D[1], false);

                for (int i = 0; i < multiTarget.textures.size(); i++) {
                    BufferedImage bimg = warp.warpImage(multiTarget.textures.get(i));
                    textures.add(bimg);
                }
            }
        }
    }

     /**
     * Create an Appearance object for use with Java3D, based on the texture
     * passed as three FloatImages
     *
     * @param red
     *            the red component of the texture to use
     * @param green
     *            the green component of the texture to use
     * @param blue
     *            the blue component of the texture to use
     * @return return a texture as an Appearance object based on the image
     *         components passed
     */
    @Override
    public  Appearance getAppearance() throws InterruptedException {
        BufferedImage img = getImage();

        int p = (int) (0.5 + Math.log(img.getWidth()) / Math.log(2.0));
        int w = (int) Math.pow(2, p);
        p = (int) (0.5 + Math.log(img.getHeight()) / Math.log(2.0));
        int h = (int) Math.pow(2, p);
        // w = h = 512;
        ImageIcon ii = new ImageIcon(img);
        BufferedImage imgResized = ImageToBufferedImage(img.getScaledInstance(
                w, h, Image.SCALE_DEFAULT), ii.getImageObserver());
        Texture2D tex = new Texture2D(Texture2D.BASE_LEVEL, Texture2D.RGB,
                imgResized.getWidth(), imgResized.getHeight());
        ImageComponent2D ic = new ImageComponent2D(ImageComponent2D.FORMAT_RGB,
                imgResized);
        tex.setMagFilter(Texture2D.NICEST);
        tex.setImage(0, ic);
        tex.setCapability(Texture.ALLOW_IMAGE_READ);
        tex.setCapability(Texture.ALLOW_IMAGE_WRITE);
        tex.setCapability(Texture.ALLOW_ENABLE_READ);
        tex.setCapability(Texture.ALLOW_ENABLE_WRITE);
        tex.setEnable(true);


        ic.setCapability(ImageComponent2D.ALLOW_IMAGE_READ);
        ic.setCapability(ImageComponent2D.ALLOW_IMAGE_WRITE);


        // Normal map
        BufferedImage nmap = this.textures.get(0); //assume 0 is a normal map
        ImageIcon ii2 = new ImageIcon(img);
        BufferedImage imgResized2 = ImageToBufferedImage(nmap.getScaledInstance(
                w, h, Image.SCALE_DEFAULT), ii2.getImageObserver());
        Texture2D tex2 = new Texture2D(Texture2D.BASE_LEVEL, Texture2D.RGB,
                imgResized2.getWidth(), imgResized2.getHeight());
        ImageComponent2D ic2 = new ImageComponent2D(ImageComponent2D.FORMAT_RGB,
                imgResized2);
        tex2.setMagFilter(Texture2D.NICEST);
        tex2.setImage(0, ic2);
        tex2.setCapability(Texture.ALLOW_IMAGE_READ);
        tex2.setCapability(Texture.ALLOW_IMAGE_WRITE);
        tex2.setCapability(Texture.ALLOW_ENABLE_READ);
        tex2.setCapability(Texture.ALLOW_ENABLE_WRITE);
        tex2.setEnable(true);


        ic2.setCapability(ImageComponent2D.ALLOW_IMAGE_READ);
        ic2.setCapability(ImageComponent2D.ALLOW_IMAGE_WRITE);

        // end normal map
        ShaderAppearance app = new ShaderAppearance();
        app.setCapability(Appearance.ALLOW_TEXTURE_READ);
        app.setCapability(Appearance.ALLOW_TEXTURE_WRITE);

        PolygonAttributes pa = new PolygonAttributes();
        pa.setCullFace(PolygonAttributes.CULL_NONE);//CULL_BACK);//
        pa.setBackFaceNormalFlip(true);
        pa.setCapability(PolygonAttributes.ALLOW_MODE_READ);
        pa.setCapability(PolygonAttributes.ALLOW_MODE_WRITE);
        app.setPolygonAttributes(pa);
        app.setCapability(Appearance.ALLOW_POINT_ATTRIBUTES_READ);
        app.setCapability(Appearance.ALLOW_POINT_ATTRIBUTES_WRITE);

        app.setCapability(Appearance.ALLOW_POLYGON_ATTRIBUTES_READ);
        app.setCapability(Appearance.ALLOW_POLYGON_ATTRIBUTES_WRITE);

        TextureAttributes ta = new TextureAttributes();
        ta.setTextureMode(TextureAttributes.MODULATE);
       // app.setTextureAttributes(ta);
       // app.setTexture(tex);
        TextureUnitState tus = new TextureUnitState();
        tus.setTexture(tex);
        tus.setTextureAttributes(ta);
       // app.setTextureUnitState(0, tus);
        TextureUnitState tus2 = new TextureUnitState();
        tus2.setTexture(tex2);
        tus2.setTextureAttributes(ta);
       // app.setTextureUnitState(1, tus2);
        TextureUnitState[] tuss = {tus, tus2};

        app.setTextureUnitState(tuss);
        
        String fragCode = "";
        String vertCode = "";

        try {
            InputStream is = MultiTextureObjectGeometry.class.getClassLoader().getResourceAsStream("factories/Bump.frag");
            Scanner sc = new Scanner(is);

            while (sc.hasNextLine()) {
                fragCode+=sc.nextLine()+"\n";
            }

            is.close();
            is = MultiTextureObjectGeometry.class.getClassLoader().getResourceAsStream("factories/Bump.vert");
            sc = new Scanner(is);
            while (sc.hasNextLine()) {
                vertCode+=sc.nextLine() + "\n";
            }

            is.close();
        } catch (IOException ex) {
            Logger.getLogger(MultiTextureObjectGeometry.class.getName()).log(Level.SEVERE, null, ex);
            return super.getAppearance();
        }

        SourceCodeShader frag = new SourceCodeShader(SourceCodeShader.SHADING_LANGUAGE_GLSL, SourceCodeShader.SHADER_TYPE_FRAGMENT, fragCode);
        SourceCodeShader vert = new SourceCodeShader(SourceCodeShader.SHADING_LANGUAGE_GLSL, SourceCodeShader.SHADER_TYPE_VERTEX, vertCode);
        GLSLShaderProgram prog = new GLSLShaderProgram();
        Shader[] shaders = {vert, frag};
        prog.setShaders(shaders);
        ShaderAttributeSet attr = new ShaderAttributeSet();
        ShaderAttributeValue subTex = new ShaderAttributeValue("subTexture", 0);
        ShaderAttributeValue mapTex = new ShaderAttributeValue("bumpTexture", 1);
        attr.put(subTex);
        
       // prog.setShaderAttrNames();
        app.setShaderProgram(prog);
        app.setShaderAttributeSet(attr);
        // Material mat = new Material();
        Material mat = new Material(new Color3f(0.5f, 0.5f, 0.5f), // ambient
                new Color3f(0.0f, 0.0f, 0.0f), // emmissive
                new Color3f(0.4f, 0.4f, 0.4f), // diffuse
                new Color3f(0.1f, 0.1f, 0.1f),// specular
                16.0f); // shine
        mat.setEmissiveColor(new Color3f(0.0f, 0.0f, 0.0f));
        mat.setLightingEnable(true);
        mat.setCapability(Material.ALLOW_COMPONENT_READ);
        mat.setCapability(Material.ALLOW_COMPONENT_WRITE);

        app.setCapability(Appearance.ALLOW_MATERIAL_READ);
        app.setCapability(Appearance.ALLOW_MATERIAL_WRITE);

        //mat.setShininess(0.0f);
        app.setMaterial(mat);
        return app;
    }


      /**
     * Returns a Java3D Shape3D object for display.
     *
     * @return This mesh as a Shape3D
     *
    public Shape3D getShape3D() throws InterruptedException {
        TriangleArray ta = getGeometry();
        Appearance app = getAppearance();
        ShaderAppearance shaderApp = new ShaderAppearance();

        Shape3D shape = new Shape3D(ta, app);
        GeometryArray g = (GeometryArray) (shape.getGeometry());
        // g = (GeometryArray)geometryDeInterleave(g);
        shape.setGeometry(g);
        shape.setBoundsAutoCompute(true);
        // mapSingleTexture(shape);
        // mapCylinderTexture(shape);

        shape.setCapability(Shape3D.ALLOW_GEOMETRY_READ);
        shape.setCapability(Shape3D.ALLOW_GEOMETRY_WRITE);
        shape.setCapability(Shape3D.ALLOW_GEOMETRY_WRITE);
        shape.setCapability(Shape3D.ALLOW_APPEARANCE_READ);
        shape.setCapability(Shape3D.ALLOW_APPEARANCE_WRITE);
        shape.setCapability(Shape3D.ALLOW_AUTO_COMPUTE_BOUNDS_READ);
        shape.setCapability(Shape3D.ALLOW_AUTO_COMPUTE_BOUNDS_WRITE);
        shape.setCapability(Shape3D.ALLOW_BOUNDS_READ);
        shape.setCapability(Shape3D.ALLOW_BOUNDS_WRITE);
        shape.setCapability(Shape3D.ENABLE_PICK_REPORTING);

        //// new
        //app.setCapability(Appearance.ALLOW_TRANSPARENCY_ATTRIBUTES_WRITE);

        g.setCapability(GeometryArray.ALLOW_COORDINATE_READ);
        g.setCapability(GeometryArray.ALLOW_COORDINATE_WRITE);
        g.setCapability(GeometryArray.ALLOW_TEXCOORD_READ);
        g.setCapability(GeometryArray.ALLOW_TEXCOORD_WRITE);
        g.setCapability(GeometryArray.ALLOW_COUNT_READ);
        g.setCapability(GeometryArray.ALLOW_REF_DATA_READ);
        g.setCapability(GeometryArray.ALLOW_REF_DATA_WRITE);
        g.setCapability(GeometryArray.ALLOW_NORMAL_READ);
        g.setCapability(GeometryArray.ALLOW_FORMAT_READ);
        // g.setCapability(GeometryArray.ALLOW_FORMAT_WRITE);
        g.setCapability(GeometryArray.ALLOW_NORMAL_WRITE);
        // g.setCapability(GeometryArray.ENABLE_PICK_REPORTING);
        // PickTool.setCapabilities(shape, PickTool.GEOMETRY_INTERSECT_INFO);
        PickTool.setCapabilities(shape, PickTool.INTERSECT_FULL);
        /*
         * app.setCapability(Appearance.ALLOW_TEXTURE_READ);
         * app.setCapability(Appearance.ALLOW_TEXTURE_WRITE);
         *
         * tex.setCapability(Texture.ALLOW_SIZE_READ);
         * /
        // tex.setCapability(Texture.ALLOW_WIDTH_WRITE);
        return shape;
    }
       * */

}
